#!/usr/bin/perl -w
## -*-cperl-*-
## CQPDemo: display larger context of match with alignment across all selected languages

use lib '.';
use My::Config;
use My::Param;
use My::KWIC;

use CGI qw/:standard *p *Tr *td *table/;

$start_match = My::Param::Get("start");
$end_match   = My::Param::Get("end");

$main_lang  = lc( My::Param::Get("language") );                                   # display main language in leftmost column
@other_lang = grep { $_ ne $main_lang and My::Param::Get("show_$_") }             # display other selected languages in following columns
  map {lc} @My::Config::Languages;
$n_columns  = 1 + @other_lang;                                                    # total number of columns
$col_width  = sprintf "%d%s", 100 / $n_columns, '%';

$n_s_context = 30;                                                                # number of sentences of context to be displayed
$s_att       = My::Config::GetAttribute("s");                                     # context defined by main language
$current_s   = $s_att->cpos2struc($start_match);

$speaker_att = My::Config::GetAttribute("speaker_name");
$tongue_att  = My::Config::GetAttribute("speaker_language");

$start_s = max( $current_s - $n_s_context, 0 );
$end_s = min( $current_s + $n_s_context + 1, $s_att->max_struc - 1 );

## print HTML header and load stylesheet
print header,
  start_html(
  -title => "CQPDemo: Extended Match Context",
  -style => { -src => "$My::Config::CSSDir/standard.css" }
  ),
  start_table( { -width => '100%',, -class => "serif", -border => 1, -rules => "all" } ),
  "\n";

## now format context for each sentence
foreach $s_num ( $start_s .. $end_s ) {
  my ( $start, $end ) = $s_att->struc2cpos($s_num);

  ## check if new speaker starts here -> display speaker's name and tongue
  my $turn = $speaker_att->cpos2struc($start);
  if ( defined $turn ) {
    my ( $turn_s, $turn_e ) = $speaker_att->struc2cpos($turn);
    if ( $turn_s == $start or $s_num == $start_s ) {
      my $spkr_name = $speaker_att->struc2str($turn) || "unknown speaker";
      my $spkr_tongue = $tongue_att->cpos2struc2str($start);
      $spkr_name .= " ($spkr_tongue)"
        if defined $spkr_tongue and $spkr_tongue ne "UNSPECIFIED";
      print Tr( td( { -colspan => $n_columns, -class => "location", -align => "center" }, escapeHTML($spkr_name) ) ), "\n";
    }
  }

  print start_Tr(), "\n";

  ## format current sentence in source language
  print start_td( { -width => $col_width, -valign => "top", -style => "padding:2px;" } );
  if ( $start_match <= $end and $end_match >= $start ) {
    print My::KWIC::Format( $main_lang, $start, $start_match - 1 ), " "
      if $start_match > $start;
    print a( { -name => "match", -class => "match" }, My::KWIC::Format( $main_lang, max( $start_match, $start ), min( $end_match, $end ) ) );
    print " ", My::KWIC::Format( $main_lang, $end_match + 1, $end )
      if $end_match < $end;
  }
  else {
    print My::KWIC::Format( $main_lang, $start, $end );
  }
  print end_td(), "\n";

  ## format aligned region in target language
  foreach $lang (@other_lang) {
    my $align_att    = My::Config::GetAttribute( lc($My::Config::CorpusBase) . $lang );
    my $align_region = $align_att->cpos2alg($start);                                      # code expects alignment to be based on sentence grid
    if ( defined $align_region ) {
      my ( $s1, $e1, $s2, $e2 ) = $align_att->alg2cpos($align_region);
      if ( $s1 == $start ) {
        ## source sentence is start of alignment bead -> display aligned region here
        my $bead_type = 0;                                                                # work out how many sentences the source part of this alignment bead contains
        my $n         = $s_num;
        while (1) {
          my ( $tmp_start, $tmp_end ) = $s_att->struc2cpos($n);
          $bead_type++;
          last if $tmp_end >= $e1;
          $n++;
        }
        print td( { -width => $col_width, -valign => "top", -rowspan => $bead_type, -style => "padding:2px;" }, My::KWIC::Format( $lang, $s2, $e2 ) ), "\n";
      }
      else {
        ## source sentence is continuation in alignment bead -> skip this cell (filled by multi-cell entry from above)
      }
    }
    else {
      ## no alignment found -> display message
      print td( { -width => $col_width, -valign => "center", -align => "center", -style => "padding:2px;" }, i( { -class => "red" }, "not translated" ) );
    }
  }
  print end_Tr(), "\n";
}

## end of HTML page
print end_html, "\n";

exit 0;

## utility functions: min(...), max(...)
sub min {
  return undef unless @_;
  my $min = shift;
  foreach (@_) { $min = $_ if $_ < $min }
  return $min;
}

sub max {
  return undef unless @_;
  my $max = shift;
  foreach (@_) { $max = $_ if $_ > $max }
  return $max;
}

