#!/usr/bin/perl -w
## -*-cperl-*-
## CQPDemo: main CGI script -- executes query and displays kwic view, frequency breakdown, or metadata distribution

use lib '.';
use My::Config;
use My::Param;
use My::KWIC;
use My::Forms;
use My::Query;

use CGI qw/:standard *table *Tr *td/;

$space = ' &nbsp;&nbsp; ';

$cqp           = $My::Config::CQP;
$action        = lc( My::Param::Get("action") );
$main_language = lc( My::Param::Get("language") );

# if activated with distribution button, dispatch to handler below
if ( $action =~ /distribution/ ) {
  do_distribution();
  exit 0;
}

# if activated with frequencies button, dispatch to handler below
elsif ( $action =~ /frequenc(y|ies)/ ) {
  do_frequencies();
  exit 0;
}

##
##  normal operation: KWIC display
##
($id, undef, $cqp_code) = My::Query::Auto(); # in CEQL mode, display generated CQP code at bottom
($size) = $cqp->exec("size $id");

print header,
  start_html(
  -title => "CQPDemo: Query Results",
  -style => { -src => "$My::Config::CSSDir/standard.css" }
  ),
  "\n";

if ( $size == 0 ) {
  print p( b("No matches.") ), end_html;
  exit 0;
}

$first = My::Param::Get("first");    # first match to display
My::Config::Error( "Match out of range.", "Sorry, match #" . ( $first + 1 ) . " does not exist." )
  if $first >= $size;
$last = $first + 19;                 # always show 20 matches on each page
$last = $size - 1
  if $last >= $size;
if ( $size >= $My::Config::MaxMatches and $first == 0 ) {
  print span( { -class => "brightred narrow" }, b("Warning: "), "Query has been limited to first $size matches!" ), br(), "\n";
}

$navbars = My::Param::Get("navbar");    # where to display navbars ($navbars contains chars "t" for top and "b" for bottom)

My::Param::Set( "show_" . $main_language, 1 );    # always display the main language (which has been queried)

if ( $navbars =~ /t/ ) {
  print p(
    table(
      { -width => '100%' },
      Tr(
        td( { -valign => "center", -align => "left", -style => "white-space:nowrap" }, My::Forms::Navbar( $first, $size ) ),
        td( { -width => '10%', -valign => "center", -align => "center", -style => "white-space:nowrap" }, b("[$size matches]") ),
        td( { -valign => "center", -align => "right" },
            start_form( -method => "POST", -action => "display.pl" ),
            My::Forms::Hidden(qw<mode language query subquery casefold sort sort_fold keyword* navbar first>),
            My::Forms::DisplayOptions(),
            $space, submit( -name => 'action', -value => 'Apply' ), end_form,
          ),
      )
    )
    ),
    "\n";
}

# display table of matches
print start_table, "\n";

$dateAtt    = My::Config::GetAttribute("text_date");
$speakerAtt = My::Config::GetAttribute("speaker_name");
$tongueAtt  = My::Config::GetAttribute("speaker_language");

@ranges = $cqp->dump( $id, $first, $last );
for $n ( $first .. $last ) {
  $idx = $n - $first;
  ( $s, $e, $t, $k ) = @{ $ranges[$idx] };
  $date = $dateAtt->cpos2struc2str($s) || "<i>unknown date</i>";
  if ( $date =~ /^([0-9]{4})-([0-9]{1,2})-([0-9]{1,2})$/ ) {
    $date = sprintf "%02d.%02d.%04d", $3, $2, $1;
  }
  $speaker = $speakerAtt->cpos2struc2str($s);
  $speaker = ($speaker) ? escapeHTML($speaker) : "<i>unknown speaker</i>";
  $tongue  = $tongueAtt->cpos2struc2str($s);
  $tongue  = ( $tongue and $tongue ne "UNSPECIFIED" ) ? " (" . escapeHTML("$tongue") . ")" : "";

  print Tr( { -class => "infoline" }, td( { -align => "right" }, b( ( $n + 1 ) . "." ) ), td( { -class => "location" }, "$speaker$tongue, $date" ) ), "\n";
  my $get_string = My::Forms::MakeGETString( [qw(language pos show_*)], { start => $s, end => $e } );
  print start_Tr, td(
    { -valign => "top", -class => "narrow" },
    a(
      {
        -href   => "context.pl?${get_string}#match",    # GET query to display context (and scroll down to match)
        -target => "context"
      },
      "context"
    )
  );
  ( $cs, $ce, $ellip_start, $ellip_end ) = My::KWIC::Context( $main_language, $s, $e );    # compute context
  print start_td( { -valign => "top", -class => "serif" } );
  print " ... " if $ellip_start;
  print My::KWIC::Format( $main_language, $cs, $s - 1, $t, $k ), " ";
  print span( { -class => "match" }, My::KWIC::Format( $main_language, $s, $e, $t, $k ) ), " ";
  print My::KWIC::Format( $main_language, $e + 1, $ce, $t, $k ), " ";
  print " ... " if $ellip_end;
  print end_td;
  print end_Tr, "\n";
  my @disp_lang = grep { $_ ne $main_language and My::Param::Get("show_$_") } map {lc} @My::Config::Languages;

  if ( @disp_lang > 0 ) {
    my $col_width = sprintf( "%d%s", 100 / @disp_lang, '%' );
    print start_Tr, td, start_td( { -valign => "top", -class => "serif" } ), start_table( { -width => '100%', -border => 1, -rules => "cols" } ), start_Tr, "\n";
    foreach my $lang (@disp_lang) {
      print start_td( { -width => $col_width, -valign => "top", -style => "padding:2px;" } );
      my ( $t_s, $t_e ) = My::KWIC::Translate( $s, $e, $lang );
      if ( not defined $t_e ) {
        print i( { -class => "red" }, "not translated" );
      }
      else {
        print My::KWIC::Format( $lang, $t_s, $t_e );
      }
      print end_td, "\n";
    }
    print end_Tr, end_table, end_td, end_Tr, "\n";
  }
}

print end_table, "\n";

if ( $navbars =~ /b/ ) {
  print p(
    table(
      { -width => '100%' },
      Tr(
        td( { -valign => "center", -align => "left", -style => "white-space:nowrap" }, My::Forms::Navbar( $first, $size ) ),
        td( { -width => '10%', -valign => "center", -align => "center", -style => "white-space:nowrap" }, b("[$size matches]") ),
        td( { -valign => "center", -align => "right" },
            start_form( -method => "POST", -action => "display.pl" ),
            My::Forms::Hidden(qw<mode language query subquery casefold sort sort_fold keyword* navbar first>),
            My::Forms::DisplayOptions(),
            $space, submit( -name => 'action', -value => 'Apply' ), end_form,
          ),
      )
    )
    ),
    "\n";
}

if ($cqp_code) {
  print p({-class => "footnote"}, "CQP query: ", code(escapeHTML($cqp_code))), "\n";
}

print end_html, "\n";

exit 0;

##
##  show distribution of matches across corpus (graphically)
##
sub do_distribution {
  my $id = My::Query::Auto();
  my ($size) = $cqp->exec("size $id");

  print header,
    start_html(
    -title => "CQPDemo: Distribution of Query Results",
    -style => { -src => "$My::Config::CSSDir/standard.css" }
    ),
    "\n";

  print p( b("$size matches."), "(Click on blue items to display examples.)" ), "\n";
  print start_table( { -align => 'left', -width => '100%' } ), start_Tr, start_td( { -valign => "top", -align => "left", -width => '50%' } ), "\n";
  print h2( { -align => "center" }, "Year" ), "\n";    # distribution for year of sitting
  my @lines      = $cqp->exec_rows("group $id match text_year");
  my @years      = qw<1996 1997 1998 1999 2000 2001 2002 2003 2004 2005 2006>;
  my %F          = map { $_ => 0 } @years;                        # frequency distribution across years
  my $total_f    = 0;
  my %Size       = map { $_ => 0 } @years;                        # number of tokens for each year
  my $total_size = 0;
  foreach my $line (@lines) {
    my ( $m, $f ) = @$line;
    next if $m eq "";
    $F{$m}   += $f;
    $total_f += $f;
  }
  my $year_att = My::Config::GetAttribute("text_year");
  foreach my $i ( 0 .. $year_att->max_struc - 1 ) {
    my ( $start, $end ) = $year_att->struc2cpos($i);
    my $m = $year_att->struc2str($i);
    $Size{$m}   += $end - $start + 1;
    $total_size += $end - $start + 1;
  }
  if ( $total_f == 0 ) {
    print p( b("No data.") );
  }
  else {
    my %Relative = ();    # pre-calculate relative frequencies (wrt. average rate), so bar sizes can be controlled
    foreach my $m (@years) {
      my $rel_size = $Size{$m} / $total_size;
      my $avg_f    = $rel_size * $total_f;      # "expected" average frequency per year
      $Relative{$m} = $F{$m} / $avg_f;
    }
    my $max_relative = max( values %Relative );
    ## default scale: 10 pixels area correspond to one occurrence
    ##  - total height of bars is 100 pixels (each bar height = relative size of year as percentage)
    ##  - for uniform distributions, all bars have length $scale (since $relative == 1)
    ##  - so total area is $scale * 100, corresponding to $scale * 10 occurrences, which must be equal to $total_f
    my $scale = $total_f / 10;
    $scale = 10                                 # lower boundary on scale, so differences remain visible
      if $scale < 5;
    $scale /= 2                                 # longest bar should not exceed 400 pixels
      while ( $max_relative * $scale > 400 );

    print start_table( { -align => "left" } ), "\n";
    foreach my $m (@years) {
      my $f        = $F{$m};
      my $relative = $Relative{$m};
      my $perc     = sprintf "%4.1f%s", 100 * $f / $total_f, '%';
      my $rel_perc = sprintf "%3.0f%s", 100 * $relative, '%';
      my $w        = $scale * $relative;
      my $h        = 100 * $Size{$m} / $total_size;
      $get_url = My::Forms::MakeGETString( [qw<mode language query casefold sort sort_fold attribute keyword* context pos show_*>],
        { action => 'show', subquery => "<match> [_.text_year='$m' \%l] !" } );
      print Tr(
        td( { -align => "left", -valign => "middle", -style => "white-space:nowrap" }, a( { -href => "display.pl?$get_url" }, b( escapeHTML($m) ) ) ),
        td( { -align => "right", -valign => "middle", -style => "white-space:nowrap" }, $perc ),
        td( { -align => "right", -valign => "middle", -style => "white-space:nowrap" }, "f=$f" ),
        td(
          { -align => "left", -valign => "middle" },
          img( { -src => "$My::Config::ImgDir/orange.gif", -width => $w, -height => $h, -align => "middle" } ),
          span( { -class => "blue", -style => "font-style:italic; font-size:90%;" }, $rel_perc ),
        ),
        ),
        "\n";
    }
    print end_table, "\n";
  }

  print end_td(), "\n", start_td( { -valign => "top", -align => "left", -width => '50%' } ), "\n";

  print h2( { -align => "center" }, "Language" ), "\n";    # distribution for speaker language
  @lines = $cqp->exec_rows("group $id match speaker_language");
  %F = ();
  foreach my $line (@lines) {
    my ( $lang, $f ) = @$line;
    next if $lang eq "";
    $F{$lang} += $f;
  }
  $total_f = 0;
  $max_f   = 0;
  foreach my $f ( values %F ) {
    $total_f += $f;
    $max_f = $f
      if $f > $max_f;
  }
  if ( $total_f == 0 ) {
    print p( b("No data.") );
  }
  else {
    my $scale = 1200 / $total_f;    # scale to percentages (100% = 1200 pixels)
    $scale = 5                      # avoid excessive stretching for low frequencies
      if $scale > 5;
    $scale /= 2                     # longest bar should not exceed 400 pixels
      while ( $max_f * $scale > 400 );

    print start_table( { -align => "left" } ), "\n";
    foreach my $lang ( sort keys %F ) {
      my $f          = $F{$lang};
      my $perc       = sprintf "%4.1f%s", 100 * $f / $size, '%';
      my $w          = $scale * $f;
      my $bar_color  = ( $lang eq "UNSPECIFIED" ) ? "grey" : "orange";
      my $print_lang = ( $lang eq "UNSPECIFIED" ) ? My::Param::Get("language") . "(?)" : $lang;
      $get_url = My::Forms::MakeGETString( [qw<mode language query casefold sort sort_fold attribute keyword* context pos show_*>],
        { action => 'show', subquery => "<match> [_.speaker_language = '$lang' \%l] !" } );
      print Tr(
        td( { -align => "left", -valign => "center", -style => "white-space:nowrap" }, a( { -href => "display.pl?$get_url" }, b( escapeHTML($print_lang) ) ) ),
        td( { -align => "right", -valign => "center", -style => "white-space:nowrap" }, "$perc" ),
        td( { -align => "right", -valign => "center", -style => "white-space:nowrap" }, "f=$f" ),
        td( { -align => "left", -valign => "center" }, img( { -src => "$My::Config::ImgDir/$bar_color.gif", -width => $w, -height => 12 } ) ),
        ),
        "\n";
    }
    print end_table, "\n";

  }

  print end_td(), end_Tr, end_table, "\n", end_html, "\n";
}

# $max = max(@list);
#   a small but really useful helper function
sub max {
  return 0
    unless @_;
  my $max = $_[0];
  map { $max = $_ if $_ > $max } @_;
  return $max;
}

##
##  compute frequency distribution of matching strings
##  (might be upgraded to an HTML form offering a wide range of different frequency tables => frequencies.pl)
##
sub do_frequencies {
  $sort = My::Param::Get("sort");    # use sort parameter to indicate desired sorting of strings
  if ( $sort eq "none" or $sort eq "random" ) {
    My::Param::Set( "sort", "asc" );    # tweak sort parameter to get correct sort clause for count command
  }
  ( $id, $sort_clause ) = My::Query::Auto();
  ($size) = $cqp->exec("size $id");

  my @lines = $cqp->exec_rows("count $id $sort_clause");    # count frequencies of matching strings (using same sort clause as in query)
  my $N = @lines;

  if ( $sort =~ /^(asc|rev)$/ ) {                  # sort matching strings (default is by frequency for sort="none")
    @lines = sort { $a->[2] cmp $b->[2] } @lines;
  }
  elsif ( $sort =~ /^(desc|drev)$/ ) {
    @lines = sort { $b->[2] cmp $a->[2] } @lines;
  }

  my $reverse = ( $sort_clause =~ /reverse/ ) ? 1 : 0;    # whether matching strings are printed reverse (we want to undo this)
  my $alignment = ($reverse) ? "right" : "left";          # whether to align matchings strings left (normal) or right (reverse sort)

  print header,
    start_html(
    -title => "CQPDemo: Frequency Table",
    -style => { -src => "$My::Config::CSSDir/standard.css" }
    ),
    "\n";

  if ( $N > 1000 ) {
    print p( b("$N unique strings found among $size matches."), "(Click on blue items to display examples.)", br "(showing only first 1000 strings)" ), "\n";
    splice @lines, 1000;
  }
  else {
    print p( b("$N unique strings found among $size matches."), "(Click on blue items to display examples.)" ), "\n";
  }

  print start_table, Tr( th( { -align => "right", -width => 200 }, "f" ), th('&nbsp;'), th( { -align => "left" }, "matching string" ) ), "\n";
  foreach my $line (@lines) {
    my ( $f, $from, $string ) = @$line;
    my $to = $from + $f - 1;
    $url = My::Forms::MakeGETString( [qw<mode language query casefold sort sort_fold attribute keyword* context pos show_*>], { action => 'show', first => $from, navbar => "b" } );
    $string = reverse($string)
      if $reverse;
    print Tr(
      td( { -align => "right" }, $f ),
      td('&nbsp;'), td( { -align => $alignment, -class => "serif" }, a( { -href => "display.pl?$url", -target => "context" }, escapeHTML($string) ) ),
      ),
      "\n";
  }
  print end_table, end_html, "\n";
}
